/*
 * Ethernet driver for Open Ethernet Controller (www.opencores.org).
 *      Copyright (c) 2002 Simon Srot (simons@opencores.org)
 *
 * Based on:
 *
 * Ethernet driver for Motorola MPC8xx.
 *      Copyright (c) 1997 Dan Malek (dmalek@jlc.net)
 *
 * mcen302.c: A Linux network driver for Mototrola 68EN302 MCU
 *
 *      Copyright (C) 1999 Aplio S.A. Written by Vadim Lebedev
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/init.h>

#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/pgtable.h>
#include <asm/bitops.h>

#include <linux/mii.h>
#include <asm-arm/arch-ep93xx/regmap.h>
#include "open_eth.h"


#define __clear_user(add,len) memset((add),0,(len))
#define _print printk
#undef DEBUG 

static const char* cardname = "ts7300-openeth";

/* The transmitter timeout
 */
#define TX_TIMEOUT	(2*HZ)

/* Buffer number
 */
#define OETH_TXBD_NUM		1
#define OETH_RXBD_NUM		4

/* Buffer size 
 */
#define OETH_TX_BUFF_SIZE	OETH_MAXBUF_LEN
#define OETH_RX_BUFF_SIZE	OETH_MAXBUF_LEN


/* The buffer descriptors track the ring buffers.   
 */
struct oeth_private {
	ushort	tx_next;			/* Next buffer to be sent */
	ushort	tx_last;			/* Next buffer to be checked if packet sent */
	ushort	tx_full;			/* Buffer ring fuul indicator */
	ushort	rx_cur;				/* Next buffer to be checked if packet received */

	oeth_regs	*regs;			/* Address of controller registers. */
	oeth_bd		*rx_bd_base;	/* Address of Rx BDs. */
	oeth_bd		*tx_bd_base;	/* Address of Tx BDs. */
	

	struct net_device_stats stats;
	struct mii_if_info mii_if;
};

static int oeth_open(struct net_device *dev);
static int oeth_start_xmit(struct sk_buff *skb, struct net_device *dev);
static void oeth_rx(struct net_device *dev);
static void oeth_tx(struct net_device *dev);
static void oeth_interrupt(int irq, void *dev_id, struct pt_regs *regs);
static int oeth_close(struct net_device *dev);
static struct net_device_stats *oeth_get_stats(struct net_device *dev);
static void oeth_set_multicast_list(struct net_device *dev);
static int oeth_set_mac_add(struct net_device *dev, void *addr);
static int calc_crc(char *mac_addr);

static int mdio_read(struct net_device *dev, int phy_id, int reg)
{
	oeth_regs *regs = (oeth_regs *)(dev->base_addr + OETH_REG_OFFSET);
	u32 timeout;
	int rxdata;

	timeout = 10;
	while (regs->miistatus & OETH_MIISTATUS_BUSY) {
		mdelay(1);
		if (--timeout < 0) {
			printk(KERN_ERR "%s: read_MII busy timeout!!\n", dev->name);
			return -1;
		}
	} 

	regs->miiaddress = (((reg<<8)&OETH_MIIADDRESS_RGAD) | (phy_id&OETH_MIIADDRESS_FIAD)); 
	regs->miicommand = OETH_MIICOMMAND_RSTAT;

	timeout = 10;
	do {
		mdelay(1);
		if (--timeout < 0) {
			printk(KERN_ERR "%s: mdio_read busy timeout!!\n", dev->name);
			return -1;
		}
	} while (regs->miistatus & OETH_MIISTATUS_BUSY);

	rxdata = (int)regs->miirx_data;

#if DEBUG
	printk("%s: mdio_read: miiaddress 0x%x rxdata 0x%x\n", dev->name, regs->miiaddress, rxdata);
#endif
	
	return rxdata;
}

static void mdio_write(struct net_device *dev, int phy_id, int reg, u16 value)
{
	oeth_regs *regs = (oeth_regs *)(dev->base_addr + OETH_REG_OFFSET);
	u32 timeout;

	timeout = 10;
	while (regs->miistatus & OETH_MIISTATUS_BUSY) {
		mdelay(1);
		if (--timeout < 0) {
			printk(KERN_ERR "%s: mdio_write busy timeout!!\n", dev->name);
			return;
		}
	} 

	regs->miiaddress = (((reg<<8)&OETH_MIIADDRESS_RGAD) | (phy_id&OETH_MIIADDRESS_FIAD)); 
	regs->miitx_data = value;
	regs->miicommand = OETH_MIICOMMAND_WCTRLDATA;
	
	timeout = 10;
	do {
		mdelay(1);
		if (--timeout < 0) {
			printk(KERN_ERR "%s: mdio_write busy timeout!!\n", dev->name);
			return;
		}
	} while (regs->miistatus & OETH_MIISTATUS_BUSY);

#if DEBUG
	printk("%s: mdio_write: miiaddress 0x%x txdata 0x%x\n", dev->name, regs->miiaddress, value);
#endif
	
}

int ks8721_reset(struct net_device *dev, int phy_addr)
{
	s16 mii_control, timeout;
	
	mii_control = mdio_read(dev, phy_addr, MII_BMCR);
	mdio_write(dev, phy_addr, MII_BMCR, mii_control | BMCR_RESET);
	for (timeout = 1000; timeout > 0; --timeout) {
		mdelay(1);
		mii_control = mdio_read(dev, phy_addr, MII_BMCR);
		if ((mii_control & BMCR_RESET) == 0)
			break;
	}
	if (mii_control & BMCR_RESET) {
		printk(KERN_ERR "%s PHY reset timeout !\n", dev->name);
		return -1;
	}
	return 0;
}

int ks8721_init(struct net_device *dev, int phy_addr)
{
	s16 data;
	
	/* Stop auto-negotiation */
	data = mdio_read(dev, phy_addr, MII_BMCR);
	mdio_write(dev, phy_addr, MII_BMCR, data & ~BMCR_ANENABLE);

	/* Set advertisement to 10/100 and Half/Full duplex (full capabilities) */
	data = mdio_read(dev, phy_addr, MII_ADVERTISE);
	data |= ADVERTISE_CSMA | ADVERTISE_100HALF | ADVERTISE_100FULL | ADVERTISE_10FULL | ADVERTISE_10HALF;
	mdio_write(dev, phy_addr, MII_ADVERTISE, data);
	
	/* Restart auto-negotiation */
	data = mdio_read(dev, phy_addr, MII_BMCR);
	data |= BMCR_ANRESTART | BMCR_ANENABLE;
	mdio_write(dev, phy_addr, MII_BMCR, data);

	return 0;
}

#if DEBUG
static void
oeth_print_packet(unsigned long add, int len)
{	
	int i;
	unsigned int *buf = (unsigned int *)add;
	int n = len / 4 + ((len & 0x3) ? 1 : 0);

	_print("ipacket: add = %lx len = %d\n", add, len);
	for(i = 0; i < n; i++) {
  		if(!(i % 4))
    			_print("\n");
  		_print(" %.8x", buf[i]);
	}
	_print("\n");
}

static void
oeth_print_packet_bytes(unsigned long add, int len)
{
	int i;
	unsigned char *buf = (unsigned char *)add;

	_print("ipacket: add = %lx len = %d\n", add, len);
	for(i = 0; i < len; i++) {
  		if(!(i % 16))
    			_print("\n");
  		_print(" %.2x", buf[i]);
	}
	_print("\n");
}
#endif

static void oeth2cpu_memcpy(unsigned char *dest, unsigned int *src, int nbytes)
{
	unsigned int dat;

	while(nbytes > 3) {
		dat = *(src++);
		*(dest++) = (dat >> 24) & 0xff;
		*(dest++) = (dat >> 16) & 0xff;
		*(dest++) = (dat >> 8) & 0xff;
		*(dest++) = dat & 0xff;
		nbytes -= 4;
	}
	if (nbytes) {
		dat = *src;
		while(nbytes--) {
			*(dest++) = (dat >> 24) & 0xff;
			dat = dat << 8;
		}
	}
}

static void cpu2oeth_memcpy(unsigned int *dest, unsigned char *src, int nbytes)
{
	unsigned int dat, i;

	while(nbytes > 3) {
		dat = *(src++) << 24;
		dat |= *(src++) << 16;
		dat |= *(src++) << 8;
		dat |= *(src++);
		*(dest++) = dat;
		nbytes -= 4;
	}
	if (nbytes) {
		i = dat = 0;
		switch (nbytes & 0x3) {
		case 3:
			dat |= *(src++) << (24 - (i++ * 8));
		case 2:
			dat |= *(src++) << (24 - (i++ * 8));
		case 1:
			dat |= *(src++) << (24 - (i++ * 8));
		case 0:
		}
		*dest = dat;
	}
}

static int
oeth_open(struct net_device *dev)
{
	struct oeth_private *cep = (struct oeth_private *)dev->priv;
	oeth_regs *regs = (oeth_regs *)(dev->base_addr + OETH_REG_OFFSET);
	u32 timeout;

	timeout = 1000;
	while (timeout-- && 
	  !(mdio_read(dev, cep->mii_if.phy_id, MII_BMSR) & BMSR_ANEGCOMPLETE))
		mdelay(1);

	if (!(mdio_read(dev, cep->mii_if.phy_id, MII_BMSR)&BMSR_ANEGCOMPLETE) )
		printk(KERN_ERR "%s BMSR_ANEGCOMPLETE timeout !\n", dev->name);
	else if (mdio_read(dev, cep->mii_if.phy_id, MII_LPA)&LPA_DUPLEX)
		regs->moder |= OETH_MODER_FULLD;

	/* Install our interrupt handler.
	 */
	request_irq(IRQ_ETH_0, oeth_interrupt, SA_SHIRQ, cardname, (void *)dev);

	/* Enable receiver and transmiter 
	 */
	regs->moder |= OETH_MODER_RXEN | OETH_MODER_TXEN;
	
	netif_start_queue(dev);

	return 0;
}

static int
oeth_close(struct net_device *dev)
{
	struct oeth_private *cep = (struct oeth_private *)dev->priv;
	oeth_regs *regs = (oeth_regs *)(dev->base_addr + OETH_REG_OFFSET);
	volatile oeth_bd *bdp;
	int i;

	/* Free interrupt hadler 
	 */
	free_irq(IRQ_ETH_0, (void *)dev);

	/* Disable receiver and transmitesr 
	 */
	regs->moder &= ~(OETH_MODER_RXEN | OETH_MODER_TXEN);	

	bdp = cep->rx_bd_base;
	for (i = 0; i < OETH_RXBD_NUM; i++) {
		bdp->len_status &= ~(OETH_TX_BD_STATS | OETH_TX_BD_READY);
		bdp++;
	}

	bdp = cep->tx_bd_base;
	for (i = 0; i < OETH_TXBD_NUM; i++) {
		bdp->len_status &= ~(OETH_RX_BD_STATS | OETH_RX_BD_EMPTY);
		bdp++;
	}

	return 0;
}

static int
oeth_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct oeth_private *cep = (struct oeth_private *)dev->priv;
	oeth_regs *regs = (oeth_regs *)(dev->base_addr + OETH_REG_OFFSET);
	volatile oeth_bd *bdp;
	unsigned long flags, lstat;

	save_flags(flags);
	cli();

	/* Fill in a Tx ring entry 
	 */
	bdp = cep->tx_bd_base + cep->tx_next;

	if (cep->tx_full) {
		oeth_tx(dev);
		if (cep->tx_full) netif_stop_queue(dev);
		return 1;
	}


	/* Copy data in preallocated buffer */
	if (skb->len > OETH_TX_BUFF_SIZE) {
		printk("%s: tx frame too long!.\n", dev->name);
		return 1; //?
	}
	else
		cpu2oeth_memcpy((int *)(dev->base_addr + bdp->addr), skb->data, skb->len);

#if DEBUG
	_print("TX\n");
	oeth_print_packet((unsigned long)(dev->base_addr + bdp->addr), skb->len);
	oeth_print_packet_bytes((unsigned long)(skb->data), skb->len);
#endif

	lstat = (skb->len << 16);
	
	/* If the frame is short, tell CPM to pad it.
	 */
	if (skb->len <= ETH_ZLEN) {
		lstat |= OETH_TX_BD_PAD;
	}

	// printk("tx #%d, status: %x\n", cep->tx_next, lstat);

	cep->tx_next = (cep->tx_next+1 == OETH_TXBD_NUM) ? 0 : (cep->tx_next+1);
	
	if (cep->tx_next == cep->tx_last) {
		cep->tx_full = 1;
	}

	/* Send it on its way.  Tell controller its ready, interrupt when done,
	 * and to put the CRC on the end.
	 */
	lstat |= (OETH_TX_BD_READY | OETH_TX_BD_IRQ | OETH_TX_BD_CRC);
	if (cep->tx_next == 0) lstat |= OETH_TX_BD_WRAP;
	bdp->len_status = lstat;

	restore_flags(flags);
	dev_kfree_skb(skb);

	return 0;
}

/* The interrupt handler.
 */
static void
oeth_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct	net_device *dev = dev_id;
	volatile struct	oeth_private *cep;
	uint int_events;
	unsigned long flags;

	cep = (struct oeth_private *)dev->priv;

	/* Get the interrupt events that caused us to be here.
	 */

	save_flags(flags);
	cli();

	/* Handle receive event in its own function.
	 */
	int_events = cep->regs->int_src;

restart_irq:
	cep->regs->int_src = int_events;
	if (int_events & (OETH_INT_RXF | OETH_INT_RXE)) 
		oeth_rx(dev_id);

	/* Handle transmit event in its own function.
	 */

	/* Check for receive busy, i.e. packets coming but no place to
	 * put them. 
	 */
	if (int_events & OETH_INT_BUSY) {
		if (!(int_events & (OETH_INT_RXF | OETH_INT_RXE)))
			oeth_rx(dev_id);
	}

	/*
	if (int_events & (OETH_INT_TXB|OETH_INT_TXC|OETH_INT_TXE)) 
		printk("tx intr: %x\n", int_events);
	*/
	oeth_tx(dev_id);

	int_events = cep->regs->int_src;
	if (int_events) goto restart_irq;
	
	restore_flags(flags);

	return;
}


static void
oeth_tx(struct net_device *dev)
{
	struct	oeth_private *cep;
	volatile oeth_bd *bdp;

	cep = (struct oeth_private *)dev->priv;

#if DEBUG
	printk("%s: oeth_tx call: tx_last 0x%x\n", dev->name, cep->tx_last);
#endif

	for (;; cep->tx_last = (cep->tx_last+1 == OETH_TXBD_NUM) ? 0 : (cep->tx_last+1)) {

		unsigned int lstat;
		bdp = cep->tx_bd_base + cep->tx_last;
		lstat = bdp->len_status;

		if ((lstat & OETH_TX_BD_READY) || 
		  ((cep->tx_last == cep->tx_next) && !cep->tx_full)) break;

		// printk("tx complete: #%d, status %x\n", cep->tx_last, lstat);
		if (cep->tx_full) {
			netif_wake_queue(dev);
			cep->tx_full = 0;
		}

		/* Check status for errors
		 */
		if (lstat & OETH_TX_BD_LATECOL)
			cep->stats.tx_window_errors++;
		if (lstat & OETH_TX_BD_RETLIM)
			cep->stats.tx_aborted_errors++;
		if (lstat & OETH_TX_BD_UNDERRUN)
			cep->stats.tx_fifo_errors++;
		if (lstat & OETH_TX_BD_CARRIER)
			cep->stats.tx_carrier_errors++;
		if (lstat & (OETH_TX_BD_LATECOL | OETH_TX_BD_RETLIM | OETH_TX_BD_UNDERRUN))
			cep->stats.tx_errors++;

		cep->stats.tx_packets++;
		cep->stats.collisions += (bdp->len_status >> 4) & 0x000f;
		// bdp->len_status = 0x0;
	}
}

static void
oeth_rx(struct net_device *dev)
{
	struct	oeth_private *cep;
	volatile oeth_bd *bdp;
	struct	sk_buff *skb;
	int	pkt_len, i;
	int	bad = 0;

	cep = (struct oeth_private *)dev->priv;

#if DEBUG
	printk("%s: oeth_rx call: rxcur 0x%x\n", dev->name, cep->rx_cur);
#endif

	/* First, grab all of the stats for the incoming packet.
	 * These get messed up if we get called due to a busy condition.
	 */
	for (i = 0; i < OETH_RXBD_NUM; i++) {

		bdp = cep->rx_bd_base + i;

#if DEBUG
		printk("%s: oeth_rx loop: bdp 0x%x len_stat 0x%x rx_cur 0x%x\n", dev->name, bdp, bdp->len_status, cep->rx_cur);
#endif

		if (bdp->len_status & OETH_RX_BD_EMPTY)
			continue;
			
		/* Check status for errors.
		 */
		if (bdp->len_status & (OETH_RX_BD_TOOLONG | OETH_RX_BD_SHORT)) {
			cep->stats.rx_length_errors++;
			bad = 1;
		}
		if (bdp->len_status & OETH_RX_BD_DRIBBLE) {
			cep->stats.rx_frame_errors++;
			bad = 1;
		}
		if (bdp->len_status & OETH_RX_BD_CRCERR) {
			cep->stats.rx_crc_errors++;
			bad = 1;
		}
		if (bdp->len_status & OETH_RX_BD_OVERRUN) {
			cep->stats.rx_crc_errors++;
			bad = 1;
		}
		if (bdp->len_status & OETH_RX_BD_MISS) {

		}
		
		if (bdp->len_status & OETH_RX_BD_LATECOL) {
			cep->stats.rx_frame_errors++;
			bad = 1;
		}
		
		
		if (bad) {

#if DEBUG
			printk("%s: oeth_rx bad: bdp 0x%x len_stat 0x%x\n", dev->name, bdp, bdp->len_status);
#endif
			bdp->len_status &= ~OETH_RX_BD_STATS;
      		bdp->len_status |= OETH_RX_BD_EMPTY;

			continue;
		}

		/* Process the incoming frame.
		 */
		pkt_len = bdp->len_status >> 16;
        
		skb = dev_alloc_skb(pkt_len + 5);
        
		if (skb == NULL) {
			printk("%s: Memory squeeze, dropping packet.\n", dev->name);
			cep->stats.rx_dropped++;
		}
		else {
			skb_reserve(skb, 2);   /*odd 16 bit alignment to make protocal stack happy*/
			skb->dev = dev;
			oeth2cpu_memcpy(skb_put(skb, pkt_len), (unsigned int *)(dev->base_addr + bdp->addr), pkt_len);
#if DEBUG
			_print("RX\n");
			oeth_print_packet((unsigned long)(dev->base_addr + bdp->addr), pkt_len);
			oeth_print_packet_bytes((unsigned long)(skb->data), skb->len);
#endif
			skb->protocol = eth_type_trans(skb,dev);
			netif_rx(skb);
			cep->stats.rx_packets++;
		}

		bdp->len_status &= ~OETH_RX_BD_STATS;
		bdp->len_status |= OETH_RX_BD_EMPTY;
	}
}

static int calc_crc(char *mac_addr)
{
	int result = 0;
	return (result & 0x3f);
}

static struct net_device_stats *oeth_get_stats(struct net_device *dev)
{
	struct oeth_private *cep = (struct oeth_private *)dev->priv;
	return &cep->stats;
}

static void oeth_set_multicast_list(struct net_device *dev)
{
	struct	oeth_private *cep;
	struct	dev_mc_list *dmi;
	volatile oeth_regs *regs;
	int	i;

	cep = (struct oeth_private *)dev->priv;

	/* Get pointer of controller registers.
	 */
	regs = (oeth_regs *)(dev->base_addr + OETH_REG_OFFSET);

	if (dev->flags & IFF_PROMISC) {
	  
		/* Log any net taps. 
		 */
		printk("%s: Promiscuous mode enabled.\n", dev->name);
		regs->moder |= OETH_MODER_PRO;
	} else {

		regs->moder &= ~OETH_MODER_PRO;
		
		if (dev->flags & IFF_ALLMULTI) {

			/* Catch all multicast addresses, so set the
			 * filter to all 1's.
			 */
			regs->hash_addr0 = 0xffffffff;
			regs->hash_addr1 = 0xffffffff;
		}
		else if (dev->mc_count) {

			/* Clear filter and add the addresses in the list.
			 */
			regs->hash_addr0 = 0x00000000;
			regs->hash_addr0 = 0x00000000;

			dmi = dev->mc_list;

			for (i = 0; i < dev->mc_count; i++) {
				
				int hash_b;

				/* Only support group multicast for now.
				 */
				if (!(dmi->dmi_addr[0] & 1))
					continue;

				hash_b = calc_crc(dmi->dmi_addr); 
				if(hash_b >= 32)
					regs->hash_addr1 |= 1 << (hash_b - 32);
				else
					regs->hash_addr0 |= 1 << hash_b;
					
			}
		}
	}
}

static int oeth_set_mac_add(struct net_device *dev, void *p)
{
	struct sockaddr *addr=p;
	volatile oeth_regs *regs;

	memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);

	regs = (oeth_regs *)(dev->base_addr + OETH_REG_OFFSET);

	regs->mac_addr1 = addr->sa_data[0] << 8 | 
		addr->sa_data[1];
    regs->mac_addr0 = 	addr->sa_data[2] << 24 	| 
		addr->sa_data[3] << 16 	| 
		addr->sa_data[4] << 8 	| 
		addr->sa_data[5];		
		
	return 0;
}

/* Initialize the Open Ethernet MAC.
 */
static int __init oeth_probe(struct net_device *dev)
{
	struct oeth_private *cep;
	volatile oeth_regs *regs;
	volatile oeth_bd *tx_bd, *rx_bd;
	int i;
	unsigned long mem_addr;

	SET_MODULE_OWNER(dev);

	cep = (struct oeth_private *)dev->priv;

	/* Allocate a new 'dev' if needed. 
	 */
	if (dev == NULL) {
		/*
		 * Don't allocate the private data here, it is done later
		 * This makes it easier to free the memory when this driver
		 * is used as a module.
		 */
		dev = init_etherdev(0, 0);
		if (dev == NULL)
			return -ENOMEM;
	}

	/* Initialize the device structure. 
	 */
	if (dev->priv == NULL) {
		cep = (struct oeth_private *)kmalloc(sizeof(*cep), GFP_KERNEL);
		dev->priv = cep;
		if (dev->priv == NULL)
			return -ENOMEM;
	}

	__clear_user(cep, sizeof(*cep));

	if (dev->base_addr == 0x0)
		dev->base_addr = (unsigned long)__ioremap(ETH_BASE_ADD, OETH_TOTAL_SIZE, 0);
	else
		dev->base_addr = (unsigned long)__ioremap((unsigned long)dev->base_addr, OETH_TOTAL_SIZE, 0);

    if(request_region(dev->base_addr, OETH_TOTAL_SIZE, cardname) == 0) {
		printk("%s: unable to request region!\n", dev->name);
        return -EAGAIN;
    }
		
	/* Get pointer ethernet controller configuration registers.
	 */
	cep->regs = (oeth_regs *)(dev->base_addr + OETH_REG_OFFSET);
	regs = (oeth_regs *)(dev->base_addr + OETH_REG_OFFSET);
	
	/* Reset the controller.
	 */
	regs->moder = OETH_MODER_RST;	/* Reset ON */
	regs->moder &= ~OETH_MODER_RST;	/* Reset OFF */
	
 	/* Setting TXBD base to OETH_TXBD_NUM.
    */
    regs->tx_bd_num = OETH_TXBD_NUM;
	
	/* Initialize TXBD pointer
	*/
	cep->tx_bd_base = (oeth_bd *)(dev->base_addr + OETH_BD_OFFSET);
	tx_bd =  (volatile oeth_bd *)(dev->base_addr + OETH_BD_OFFSET);
	
	/* Initialize RXBD pointer
	*/
	cep->rx_bd_base = ((oeth_bd *)(dev->base_addr + OETH_BD_OFFSET)) + OETH_TXBD_NUM;
	rx_bd =  ((volatile oeth_bd *)(dev->base_addr + OETH_BD_OFFSET)) + OETH_TXBD_NUM;
	
	/* Initialize transmit pointers.
	 */
	cep->rx_cur = 0;

	/* or read it from eth register, or set it to 0 in eth */
	cep->tx_next = 0;
	cep->tx_last = 0;
	cep->tx_full = 0;

	/* Set min/max packet length
	*/
	regs->packet_len = 0x00400600;

	/* Set IPGT register to recomended value 
	 */
	regs->ipgt = 0x00000012;

	/* Set IPGR1 register to recomended value 
	 */
	regs->ipgr1 = 0x0000000c;

	/* Set IPGR2 register to recomended value 
	 */
	regs->ipgr2 = 0x00000012;

	/* Set COLLCONF register to recomended value 
	 */
	regs->collconf = 0x000f003f;
	
	/* Set control module mode 
	 */
#if 0
	regs->ctrlmoder = OETH_CTRLMODER_TXFLOW | OETH_CTRLMODER_RXFLOW;
#else
	regs->ctrlmoder = 0;
#endif

	/* Configure Micrel KS8721BL PHY
	*/
 	cep->mii_if.dev = dev;
	cep->mii_if.phy_id = MICREL_PHY;
	cep->mii_if.phy_id_mask = OETH_MIIADDRESS_FIAD;
	cep->mii_if.reg_num_mask = OETH_MIIADDRESS_RGAD;	
	cep->mii_if.mdio_read = mdio_read;
	cep->mii_if.mdio_write = mdio_write;
	
	ks8721_reset(dev, cep->mii_if.phy_id);
	ks8721_init(dev, cep->mii_if.phy_id);
  
	regs->miimoder = 0x164;
	
	/* Initialize TXBDs.
	 */
	mem_addr = OETH_BUF_OFFSET; 
	for(i = 0; i < OETH_TXBD_NUM; i++) {
		tx_bd[i].len_status = OETH_TX_BD_PAD | OETH_TX_BD_CRC | OETH_RX_BD_IRQ;
		tx_bd[i].addr = mem_addr;
		mem_addr += OETH_TX_BUFF_SIZE;
	}
	tx_bd[OETH_TXBD_NUM - 1].len_status |= OETH_TX_BD_WRAP;

	/* Initialize RXBDs.
	 */
	mem_addr = OETH_BUF_OFFSET + (OETH_TX_BUFF_SIZE*OETH_TXBD_NUM); 
	for(i = 0; i < OETH_RXBD_NUM; i++) {
		rx_bd[i].len_status = OETH_RX_BD_EMPTY | OETH_RX_BD_IRQ;
		rx_bd[i].addr = mem_addr;
		mem_addr += OETH_RX_BUFF_SIZE;
	}
	rx_bd[OETH_RXBD_NUM - 1].len_status |= OETH_RX_BD_WRAP;


	/* Set default ethernet station address to ep93xx eth addr + 1.
	 */
	{
		unsigned char *epmac = IO_ADDRESS(MAC_INDAD);
		unsigned long mac = epmac[5] | (epmac[4] << 8) | (epmac[3] << 16);

		mac++;

        	*(unsigned long *)IO_ADDRESS(MAC_AFP) = 0x00; 
		dev->dev_addr[0] = epmac[0];
		dev->dev_addr[1] = epmac[1];
		dev->dev_addr[2] = epmac[2];
		dev->dev_addr[3] = (mac >> 16) & 0xff;
		dev->dev_addr[4] = (mac >> 8) & 0xff;
		dev->dev_addr[5] = mac & 0xff;
	}

	regs->mac_addr1 = dev->dev_addr[0] << 8 | dev->dev_addr[1];
	regs->mac_addr0 = dev->dev_addr[2] << 24 | dev->dev_addr[3] << 16 | 
	  dev->dev_addr[4] << 8 | dev->dev_addr[5];
	
	/* Clear all pending interrupts 
	 */
 	regs->int_src = 0xffffffff;

	/* Promisc, IFG, CRCEn
	 */
	regs->moder |= OETH_MODER_PAD | OETH_MODER_IFG | OETH_MODER_CRCEN;

	/* Enable interrupt sources.
	 */
	regs->int_mask = OETH_INT_MASK_TXE 	| 
			OETH_INT_MASK_TXC 	| 
			OETH_INT_MASK_TXB 	| 
			OETH_INT_MASK_RXF 	| 
			OETH_INT_MASK_RXE 	|
			OETH_INT_MASK_BUSY 	|
			OETH_INT_MASK_RXC;

	/* Fill in the fields of the device structure with ethernet values. 
	 */
	ether_setup(dev);

	/* The Open Ethernet specific entries in the device structure. 
	 */
	dev->open = oeth_open;
	dev->hard_start_xmit = oeth_start_xmit;
	dev->stop = oeth_close;
	dev->get_stats = oeth_get_stats;
	dev->set_multicast_list = oeth_set_multicast_list;
	dev->set_mac_address = oeth_set_mac_add;

	
#if DEBUG
	printk("%s: Open Ethernet Memory Map:\n", dev->name);
	printk("\tRegister Base=0x%x \n\tTXBD=0x%x TXBUF=0x%x \n\tRXBD=0x%x RXBUF=0x%x\n", 
		(uint)cep->regs, (uint)cep->tx_bd_base, tx_bd[0].addr, (uint)cep->rx_bd_base, rx_bd[0].addr);
#endif
		
	printk("%s: Open Ethernet Core Version 1.0\n", dev->name);

	return 0;
}



#ifdef MODULE

static struct net_device dev_openeth = { 
	init: oeth_probe 
	};
	
static int io = ETH_BASE_ADD;
static int irq = IRQ_ETH_0;

MODULE_PARM(io, "i");
MODULE_PARM(irq, "i");

MODULE_LICENSE("GPL");

int init_module(void)
{
	int result;

	dev_openeth.base_addr = io;
	dev_openeth.irq       = irq;
	dev_openeth.init      = oeth_probe;

	if ((result = register_netdev(&dev_openeth)) != 0)
		return result;

	return 0;
}

void cleanup_module(void)
{
	free_irq(dev_openeth.irq, &dev_openeth);
	release_region(dev_openeth.base_addr, OETH_TOTAL_SIZE);
	
	if (dev_openeth.priv)
		kfree(dev_openeth.priv);

	unregister_netdev(&dev_openeth);		
}

#endif /* MODULE */




