/*
 *  linux/drivers/serial/tsuart73.c
 *
 *  Driver for Technologic Systems UART
 *
 *  (c) 2005 Technologic Systems
 *
 * This source code is hereby licensed to the Linux 2.4 kernel under the
 * GPL, and to Technologic Systems customers under the BSD license.
 *
 * 2005-11-17:mos:initial version completed
 * 2005-11-18:mos:added proper hardware flow control support (Tx)
 * 2006-01-18:mos:cleanup;now using memory access instead of inb/outb
 * 2006-01-26:mos:added support for 9-bit mode
 */
/*
  1. Add support for 9-bit serial
  -a bit in the upper byte of STATUS will enable/disable 9-bit mode
  -in 9-bit mode we take/return two bytes per write/read
  -we must make sure everything is flushed when we enter/leave this mode
   so that we never lose synchronization

   9-bit mode is implemented with a separate device (ttyT9Sx).
   The 8 and 9 bit devices for a port are prohibited from being open at
   the same time.

   Fetch data to read a byte at a time until we see a valid most
   significant byte (x & 0xFE == 0).

  2. Add support for software based parity (8+1,7+1). Tell the hardware to
  use one more data bit, then use it for parity.
 */
/*
  Implementation
  1. Implement functions to 
     (a) get current 8/9 bit mode status
     (b) set current 8/9 bit mode status
     DONE
  2. Identify where we actually need to use the above functions, and
     create a cache of each port's status so we can use it for frequent
     access.
     1. Calling TXDAT or RXDAT.  I think we might simply hide the
     8/16 bit difference in these functions, and have them track the
     state.
     DONE
  3. Identify everywhere that RXDAT is currently being used. Change
     each location to use 8/16 bits depending on the current mode.
     Modify write logic to synchronize byte-wise on msB & 0xFE == 0,
     and read logic to clear those bits.
     DONE
 */
#include <linux/slab.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/serial_core.h>
#include <asm/hardware/serial_amba.h>
#include <linux/ts_sbc.h>
#include <asm/uaccess.h>
#include <asm/delay.h>

#define DELAY_TIME      HZ / 100        // 100 Hz
#define UART_PORT_SIZE  64              // not used
#define PORT_TSUART     55              // see include/linux/serial_core.c
#define UART_NR		16
#define UART_CLOCK	14745600	// not used
#define UART_FIFOSIZE	5		// not used

#define SERIAL_TSUART_MAJOR	234	// in unassigned range (231-239)
#define SERIAL_TSUART_MINOR	1
#define SERIAL_TSUART_NR	UART_NR
#define SERIAL_TSUART_NAME	"ttyTS"
#define SERIAL_TSUART9_NAME	"ttyT9S"

#define CALLOUT_TSUART_NAME	"cuats"
#define CALLOUT_TSUART9_NAME	"cuat9s"
#define CALLOUT_TSUART_MAJOR	235
#define CALLOUT_TSUART_MINOR	1
#define CALLOUT_TSUART_NR	UART_NR

#define TS_MODEL_REG 0x2000000
#define TS_TTL_REG   0x2000000
#define TSUART_IRQ 40

#define TSUART1_BASE 0x72000000

//-----------------------
// Forward Declarations <
//-----------------------

int running[8];
static struct uart_port tsuart_ports[];
static void tsuart_stop_tx(struct uart_port *port, unsigned int tty_stop);
static struct timer_list timer;
static int timer_added = 0;
//#define TSUART_DEBUG
//#define DEBUG_IO
//#define DEBUG_IO2
//#define DEBUG_IRQ
//#define TSUART_DEBUG_DATA
//#define TSUART_STATS
//-----------------------------------------------------------------------
//                                                                      /
// Interrupt Routines                                                   \_
//                                                                      /
//                                                                      \_
//                                                                      /
//                                                                      \_
//-----------------------------------------------------------------------

/*
  TS UART interrupt

  There are up to three UARTS.  We don't know which UART needs what 
  service, but whatever needs doing we must do it *fast* because 
  theoretically we could have 6 streams of data (3 transmit and 
  3 receive) each running at 115200 baud.

  Our IRQ line will be asserted as long as there is data to receive,
  so if more data comes in after we finish looking we will just
  get another interrupt.

  On the other hand, the IRQ line will only be asserted on the edge
  when the transmit buffer becomes empty, and the empty condition
  will not re-assert the IRQ.  Further, any access to the UART register
  will clear the IRQ. This is necessary because we can't always do 
  anything in response to a transmit buffer empty condition. For example 
  we can't put data in the transmit buffer if there is none to send.  
  So we have to take extra care to not "miss" a TBRE generated IRQ; this 
  extra care consists of making sure there aren't any race conditions 
  whereby TBRE is asserted after we check the status but before we
  perform a subsequent UART access.

  Pseudocode:
    int data1=0,data2=0,data3=0,counter=256;
    while (--counter) {
      stat1 = Read STAT register of UART 1
      if receive data ready in stat1 (bit 1, DR is set)
        data1 = read RXDAT register of UART 1
      Repeat previous 3 lines, except for UARTs 2 and 3
      if data1 != 0
        push data1 to the appropriate receive location
      if data2 != 0
        push data2 to the appropriate receive location
      if data3 != 0
        push data3 to the appropriate receive location
      if (data1 + data2 + data3 == 0)
        break
    }
    -- if counter is zero we might lose data. the alterative is
    -- to not limit how much data the interrupt can read from the
    -- UARTs, but this is a bad idea as it can make the system
    -- unresponsive and will probably overflow a buffer somewhere
    -- else in the system anyway.
    stat1 = Read STAT register of UART 1
    if transmit buffer empty in stat1 (bit 0, TBRE is set)
      try putting the next byte to send to UART1
    repeat previous 3 lines, except for UARTs 2 and 3
    -- since we did all our reads already, the only possible
    -- access to the UART after reading the STAT register is to
    -- write another byte of data to the transmit buffer. Therefore
    -- there is no chance of missing TBRE assertion.
 */

//-----------------------------
// Begin hardware access code <
//-----------------------------

static inline int get_ts_model(void) {
  return (inw(TS7XXX_IO16_BASE + TS_MODEL_REG)) & 3;
}

static inline int get_tsuart_ttl(void) {
  unsigned char c;
  c = inb(TS7XXX_IO8_BASE + TS_TTL_REG);
  return c;
}

static inline void put_tsuart_ttl(int flag) {
  if (flag) {
    outb(get_tsuart_ttl() | 64,TS7XXX_IO8_BASE + TS_TTL_REG);
  } else {
    outb(get_tsuart_ttl() & (~64),TS7XXX_IO8_BASE + TS_TTL_REG);
  }
}

// read the TS UART STAT register for the specified UART
static inline unsigned short get_tsuart_STAT(struct uart_port *port) {
  volatile unsigned short *STAT = (unsigned short *)port->membase;
  unsigned short tmp;

  tmp = *STAT;
#ifdef DEBUG_IO
  printk("***STAT read: port %d, data=%X\n",port->line,tmp);
#endif
  return tmp;
}

// write the TS UART STAT register for the specified UART
static inline void put_tsuart_STAT(struct uart_port *port,unsigned short value) {
  volatile unsigned short *STAT = (unsigned short *)port->membase;

#ifdef DEBUG_IO
  printk("***STAT write: port %d, data=%X\n",port->line,value);
#endif
  *STAT = value;
}

static inline int bit_MODE9(unsigned short reg_value);
static inline unsigned short assign_bit_MODE9(unsigned short reg_value,int set);
static inline int MODE9(struct uart_port *port) {
  return (bit_MODE9(get_tsuart_STAT(port)) != 0);
}

static inline void set_MODE9(struct uart_port *port,int on) {
  put_tsuart_STAT(port,assign_bit_MODE9(get_tsuart_STAT(port),on));
}

#define rXmsB port->unused[0]
#define tXmsB port->unused[1]
#define tXdaTmsB port->unused[2]
#define rXdaTlsB port->unused[3]
#define NINER port->unused[4]
/*
 */
// read the TS UART RXDAT register for the specified UART
static inline int get_tsuart_RXDAT(struct uart_port *port) {
  volatile unsigned short *DAT = (unsigned short *)port->membase;

  if (running[port->line] == 9) {
    return DAT[1] & 0x1FF;
  } else {
    return DAT[1] & 0xFF;
  }
}

// write the TS UART TXDAT register for the specified UART
static inline void put_tsuart_TXDAT(struct uart_port *port,unsigned short data) {
  volatile unsigned short *DAT = (unsigned short *)port->membase;

#ifdef DEBUG_IO
  printk("***TX port %d, data=%X\n",port->line,data);
#endif
#ifdef DEBUG_IO2
    printk("Tx with stat=%X\n",get_tsuart_STAT(port));
#endif
  if (MODE9(port)) {
    if (tXmsB) {
#ifdef TSUART_DEBUG_DATA
      printk("tXmsb=%d,data=%X\n",tXmsB,data);
#endif
      if ((data & 0xFE) == 0) {
	tXdaTmsB = data;
	tXmsB = !tXmsB;
      } else {
#ifdef DEBUG_IO
	printk("***TX Rejected (sync)\n");
#endif
      }
    } else {
#ifdef TSUART_DEBUG_DATA
      printk("tXmsb=%d,writing %X\n",tXmsB,(((unsigned short)tXdaTmsB) << 8) + (data & 0xFF));
#endif
      DAT[1] = (((unsigned short)tXdaTmsB) << 8) + (data & 0xFF);
      tXmsB = !tXmsB;
    }
  } else {
      DAT[1] = data & 0xFF;
  }
}

static inline int get_tsuart_TRI(void) {
  return ((inw(TS7XXX_IO16_BASE + 0x2400000) & 1) == 1);
}

//--------------------------
// Begin OS dependant code <
//--------------------------


int rxof=0;
int mincounter=256;
// lets us receive multiple devices before pushing it up
int gotbytes[UART_NR],flushing[UART_NR];

// process the byte in the lsb of data by sending it to the
// appropriate OS buffers for the given TS UART UART
static void tsuart_handleRx(struct uart_port *port,int data) {
  struct tty_struct *tty;
  unsigned int ch = data;

#ifdef TSUART_DEBUG_DATA
  printk("(%d):",port-tsuart_ports);
  printk("Rx%d:%02X%s\n",port->line,ch,port->info?"":"?");
#endif
  if (!port->info || flushing[port->line]) {
    return; 
  }
  tty = port->info->tty;
  if (!port->info->tty) {
    printk("Hic!\n");
    return;
  }
  if (tty->flip.count >= TTY_FLIPBUF_SIZE) {
    tty_flip_buffer_push(tty); 
  }
  if (running[port->line] == 9) {
    tty_insert_flip_char(tty, (ch >> 8) & 1, 0);
    tty_insert_flip_char(tty, ch & 0xFF, 0);
  } else {
    tty_insert_flip_char(tty, ch, 0);
  }
  gotbytes[port->line]++;
  return; 
}

static unsigned int total_sent = 0;
static inline int bit_CTS(unsigned short reg_value);
// If there is data ready to be transmitted for the given UART,
// then feed to the byte to the UART
static void tsuart_try_handleTx(struct uart_port *port) {
  struct circ_buf *xmit;
  unsigned int ch=0;

  if (!port->info) {
    return;
  }
  if (port->info->flags & ASYNC_CTS_FLOW) {
    if (bit_CTS(get_tsuart_STAT(port))) {
      port->info->tty->hw_stopped = 0;
    } else {
      port->info->tty->hw_stopped = 1;
      if (!timer_added) {
	timer_added = 1;
	//	printk("Adding timer (%d)\n",port->line);
	timer.expires = jiffies + DELAY_TIME;
	add_timer (&timer);
      }
    }
  }
  xmit = &port->info->xmit;
  if (uart_circ_empty(xmit) || uart_tx_stopped(port)) {
#ifdef TSUART_DEBUG
    if (uart_tx_stopped(port)) {
      printk("tx stop detected\n");
    }
#endif
    tsuart_stop_tx(port,0);
    return;
  }
  // handle one character
  ch = xmit->buf[xmit->tail];
#ifdef TSUART_DEBUG_DATA
  printk("(%d):",port-tsuart_ports);
  printk("Tx%d:%02X%s\n",port->line,ch,port->info?"":"?");
#endif
  total_sent++;
  put_tsuart_TXDAT(port,(unsigned char)ch);
  xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
  port->icount.tx++;
  // done handling one character

  if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS) {
    uart_write_wakeup(port);
  }
  if (uart_circ_empty(xmit)) {
#ifdef TSUART_DEBUG
    printk("Stopping xmit at end\n");
#endif
    tsuart_stop_tx(port, 0);
  }
}


//----------------------------
// Begin FPGA dependant code <
//----------------------------

static inline int bit_MODE9(unsigned short reg_value) {
  return (reg_value & 0x100) == 0x100;
}

// accepts the value read from the TS UART STAT register
// returns TRUE if the TBRE bit is set, and FALSE if it is clear
static inline int bit_TBRE(unsigned short reg_value) {
  return (reg_value & 0x01) == 0x01;
}

// accepts the value read from the TS UART STAT register
// returns TRUE if the DR bit is set, and FALSE if it is clear
static inline int bit_DR(unsigned short reg_value) {
  return (reg_value & 0x02) == 0x02;
}

// accepts the value read from the TS UART STAT register
// returns TRUE if the OERR bit is set, and FALSE if it is clear
static inline int bit_OERR(unsigned short reg_value) {
  return (reg_value & 0x04) == 0x04;
}

// accepts the value read from the TS UART STAT register
// returns TRUE if the CTS bit is set, and FALSE if it is clear
static inline int bit_CTS(unsigned short reg_value) {
  return (reg_value & 0x08) == 0x08;
}

// accepts the value read from the TS UART STAT register
// returns TRUE if the RTS bit is set, and FALSE if it is clear
static inline int bit_RTS(unsigned short reg_value) {
  return (reg_value & 0x10) == 0x10;
}

static inline unsigned short assign_bit_MODE9(unsigned short reg_value,int set) {
  return (reg_value & ~0x100) | (set ? 0x100 : 0);
}

// accepts the value read from the TS UART STAT register
// If set is TRUE, sets the RTS bit, others clears it,
// and returns the new value to be written back to the STAT register
static inline unsigned short assign_bit_RTS(unsigned short reg_value,int set) {
  return (reg_value & ~0x10) | (set ? 0x10 : 0);
}

// accepts the value read from the TS UART STAT register
// returns the baud rate
static int bits_BAUD(unsigned short reg_value) {
  int encoded_rate;

  encoded_rate = (reg_value & 0xE0) >> 5;
  switch (encoded_rate) {
  case 0: return 115200;
  case 1: return 57600;
  case 2: return 38400;
  case 3: return 19200;
  case 4: return 9600;
  case 5: return 4800;
  case 6: return 2400;
  case 7: return 0;    // IRQ off
  }
  return 0; // not reached
}

// accepts the value read from the TS UART STAT register,
// and the encoded baud rate to set, and returns the value to be written
// back to the STAT register.
static inline unsigned short encode_BAUD(unsigned short reg_value,int bits) {
  return (reg_value & 0xFF1F) | (bits << 5);
}

// accepts the value read from the TS UART STAT register,
// encodes the given baud rate, and then returns the value to be written
// back to the STAT register.  If the passed baud rate is not supported
// the passed value is returned unchanged.
static unsigned short assign_bits_BAUD(unsigned short reg_value,int baud_rate) {
  switch (baud_rate) {
  case 115200: return encode_BAUD(reg_value,0);
  case 57600:  return encode_BAUD(reg_value,1);
  case 38400:  return encode_BAUD(reg_value,2);
  case 19200:  return encode_BAUD(reg_value,3);
  case 9600:   return encode_BAUD(reg_value,4);
  case 4800:   return encode_BAUD(reg_value,5);
  case 2400:   return encode_BAUD(reg_value,6);
  case 0:      return encode_BAUD(reg_value,7);
  default:     return reg_value;
  }
}

//-----------------------------------------------------------------------
//                                                                      /
// /proc Routines                                                       \_
//                                                                      /
//                                                                      \_
//                                                                      /
//                                                                      \_
//-----------------------------------------------------------------------
/*
  The user can set COM3 to TTL levels as follows:
    echo "1" > /proc/device/tsuart/com3ttl
  The user can set COM3 to RS-232 levels as follows:
    echo "0" > /proc/device/tsuart/com3ttl
  The user can query the current status of COM3 as follows:
    cat /proc/device/tsuart/com3ttl
      "0" if COM3 is at RS-232 levels
      "1" if COM3 is at TTL levels
 */
int procfile_read(char *buffer, 
		  char **buffer_location, 
		  off_t offset, 
		  int buffer_length, 
		  int *eof,
		  void *data)
{
  if (offset > 0) {
    *eof = 1;
    return 0;
  }
  MOD_INC_USE_COUNT;
  buffer[0] = (get_tsuart_ttl() & 64) ? '1' : '0';
  *buffer_location = buffer;
  *eof = 1;
  MOD_DEC_USE_COUNT;

  return 1;
}

int procfile_write(struct file *file,
		   const char *buffer,
		   unsigned long count,
		   void *data) {
  char tmp;

  MOD_INC_USE_COUNT;
  if (copy_from_user(&tmp,buffer,1)) {
    return -EFAULT;
  }
  if (tmp == '1') {
    put_tsuart_ttl(1);
  } else if (tmp == '0') {
    put_tsuart_ttl(0);
  }
  MOD_DEC_USE_COUNT;
  return 1;
}

int procfile_init(void) {
  struct proc_dir_entry *dir1,*file;

  dir1 = proc_mkdir("driver/tsuart",NULL);
  if (dir1 == NULL) { return -ENOMEM; }
  
  file = create_proc_entry("driver/tsuart/com3ttl",
			   S_IWUSR |S_IRUSR | S_IRGRP | S_IROTH,
			   NULL);
  if (file == NULL) {
    remove_proc_entry("driver/tsuart", NULL);
    return -ENOMEM;
  }
  file->read_proc = procfile_read;
  file->write_proc = procfile_write;
  file->owner = THIS_MODULE;
  return 0;
}

void procfile_exit(void) {
  remove_proc_entry("driver/tsuart/com3ttl",NULL);
  remove_proc_entry("driver/tsuart", NULL);
}


//---------------------
// End dependant code <
//---------------------

int tsuart_sanity_check_STAT(struct uart_port *port) {
  unsigned short stat;

  stat = get_tsuart_STAT(port);
  put_tsuart_STAT(port,stat | 0x0F);
  return (stat == get_tsuart_STAT(port));
}

int tsuart_sanity_check_baud(struct uart_port *port,int rate) {
  unsigned short stat;

  stat = get_tsuart_STAT(port);
  put_tsuart_STAT(port,assign_bits_BAUD(stat,rate));
  stat = get_tsuart_STAT(port);
  return (bits_BAUD(stat) == rate);
}

// Check if the given UART port looks sane.
int tsuart_sanity_check(struct uart_port *port) {
  if (!tsuart_sanity_check_STAT(port)) { return 0; }
  if (!tsuart_sanity_check_baud(port,2400)) { return 0; }
  if (!tsuart_sanity_check_baud(port,4800)) { return 0; }
  if (!tsuart_sanity_check_baud(port,9600)) { return 0; }
  if (!tsuart_sanity_check_baud(port,19200)) { return 0; }
  if (!tsuart_sanity_check_baud(port,38400)) { return 0; }
  if (!tsuart_sanity_check_baud(port,57600)) { return 0; }
  if (!tsuart_sanity_check_baud(port,115200)) { return 0; }
  return 1;
}

// Check if data is ready to be received from the specified UART
// and if so, read it into the data location provided.
static inline void tsuart_checkRx(struct uart_port *port,unsigned int *data) {
  int stat;

  if (!port->membase) {
    return;
  }
  stat = get_tsuart_STAT(port);

  if (bit_DR(stat)) {
    if (bit_OERR(stat)) {
      rxof++;
    }
    *data = get_tsuart_RXDAT(port);
  }
}



// See if the given UART wants to transmit data, and if so, 
// transmit data if any is ready to be sent.
static inline void tsuart_checkHandleTx(struct uart_port *port) {
  int stat;

  if (!port->membase) {
    return;
  }
  // Check the STAT register, and if it says the transmit
  // buffer is empty, then check if we have any data to send.
  // We do it in this order for two reasons. First, because if 
  // the transmit register became empty after our last access 
  // to the UART, we may wind up getting another interrupt for 
  // which we may have nothing to do (except clear the IRQ), so
  // we might as well do that here.  Second, it probably less
  // expensive to check STAT then it is to see if we have more
  // data to send, so if we don't have to do both we might as
  // well not do the more expensive operation.
  
  
  stat = get_tsuart_STAT(port);
  if (bit_TBRE(stat)) { // if buffer empty, then try to send
    tsuart_try_handleTx(port);
  }
}

// See if we read any data, and if so, pass it to the upper layers
static inline void tsuart_try_handleRx(struct uart_port *port,int data) {
  if (data >= 0) {
    tsuart_handleRx(port,data);
  }
}

#ifdef TSUART_STATS
int stats[256];
#endif

static void tsuart_int8(int irq, void *dev_id, struct pt_regs *regs) {
  struct uart_port *port = (struct uart_port *)(tsuart_ports);
  unsigned int data[8],counter=256;
  int i;

  for (i=0;i<8;i++) {
    gotbytes[i] = 0;
  }

#ifdef DEBUG_IRQ
  printk("*** int8\n");
#endif
  while (--counter) {
    for (i=0;i<8;i++) {
      data[i] = -1;
    }
    for (i=0;i<8;i++) {
      if (running[i] == 9) {
	tsuart_checkRx(port+i+8,data+i);
      } else {
	tsuart_checkRx(port+i,data+i);
      }
    }
    for (i=0;i<8;i++) {
      if (data[i] != -1) {
	break;
      }
    }
    if (i == 8) {
      break;
    }
    for (i=0;i<8;i++) {
      if (running[i] == 9) {
	tsuart_try_handleRx(port+i+8,data[i]);
      } else {
	tsuart_try_handleRx(port+i,data[i]);
      }
    }
  }
  if (counter < mincounter) { mincounter = counter; }
  for (i=0;i<8;i++) {
    if (running[i] == 9) {
      tsuart_checkHandleTx(port+i+8);
    } else {
      tsuart_checkHandleTx(port+i);
    }
  }
  for (i=0;i<8;i++) {
    if (gotbytes[i]) {
#ifdef TSUART_STATS
      stats[gotbytes[i]-1]++;
#endif
#ifdef TSUART_DEBUG_DATA
      printk("tsuart: pushing buffer %d\n",i);
#endif
      if (running[i] == 9) {
	tty_flip_buffer_push((port+i+8)->info->tty);
      } else {
	tty_flip_buffer_push((port+i)->info->tty);
      }
    }
  }
}


//-----------------------------------------------------------------------
// Interface Routines                                                   /
//                                                                      \_
//                                                                      /
//                                                                      \_
//                                                                      /
//                                                                      \_
//-----------------------------------------------------------------------

static void tsuart_stop_tx(struct uart_port *port, unsigned int tty_stop)
{
#ifdef TSUART_DEBUG
  //  printk("tsuart_stop_tx called, port=%p,tty_stop=%d\n",port,tty_stop);
#endif
}

static void tsuart_start_tx(struct uart_port *port, unsigned int tty_start)
{
#ifdef TSUART_DEBUG
  printk("tsuart_start_tx\n");
#endif
  tsuart_checkHandleTx(port);
}

static void tsuart_stop_rx(struct uart_port *port)
{
#ifdef TSUART_DEBUG
  printk("tsuart_stop_rx called, port=%p\n",port);
#endif
}

static void tsuart_enable_ms(struct uart_port *port)
{
#ifdef TSUART_DEBUG
  printk("tsuart_enable_ms called,port=%p\n",port);
#endif
}

static unsigned int tsuart_tx_empty(struct uart_port *port)
{
#ifdef TSUART_DEBUG
  printk("tsuart_tx_empty called,port=%p\n",port);
#endif
  return bit_TBRE(get_tsuart_STAT(port));
}

static unsigned int tsuart_get_mctrl(struct uart_port *port)
{
  unsigned int result = 0;

#ifdef TSUART_DEBUG
  printk("tsuart_get_mctrl called,port=%p\n",port);
#endif

  result |= TIOCM_CAR;  // pretend DCD is always set
  result |= TIOCM_DSR;  // pretend DSR is always set
  if (bit_CTS(get_tsuart_STAT(port))) {
#ifdef TSUART_DEBUG
    printk("reporting CTS asserted\n");
#endif
    result |= TIOCM_CTS;
  } else {
#ifdef TSUART_DEBUG
    printk("reporting CTS de-asserted\n");
#endif
  }
  return result;
}

static void tsuart_set_mctrl(struct uart_port *port, unsigned int mctrl)
{
  if (port->line > 3) {
    return;
  }
#ifdef TSUART_DEBUG
  printk("tsuart_set_mctrl called,port=%p,mctrl=%X\n",port,mctrl);
  printk("RTS=%d\n",((mctrl&TIOCM_RTS) != 0) ? 1:0);
#endif
  put_tsuart_STAT(port,
		  assign_bit_RTS(get_tsuart_STAT(port),(mctrl&TIOCM_RTS)!=0));
  // No support for TIOCM_DTR, so ignore it
}

static void tsuart_break_ctl(struct uart_port *port, int break_state)
{
  // UART does not support BREAK
#ifdef TSUART_DEBUG
  printk("tsuart_break_ctl called, port=%p, bs=%d\n",port,break_state);
#endif
}

static int started = 0;

static void tsuart_cts_timer(unsigned long data) {
  struct uart_port *port = (struct uart_port *)(tsuart_ports);
  //  struct uart_port *port = (struct uart_port *)data;
  //  printk("timer fired!\n");
  timer_added = 0;
  tsuart_checkHandleTx(port);
  tsuart_checkHandleTx(port+1);
  tsuart_checkHandleTx(port+2);
}

static int tsuart_startup(struct uart_port *port)
{
#ifdef TSUART_STATS
  int i;
#endif
  int tmp;

  if (running[port->line]) {
    return -EBUSY; // already open!
  }
  running[port->line] = NINER ? 9 : 8;
  if (started++) {
    return 0;
  }
  set_MODE9(port,NINER?1:0);

  /* create our timer but don't submit it just yet */
  timer.data = (unsigned long )port;
  timer.function = tsuart_cts_timer;

#ifdef TSUART_STATS
  for (i=0;i<256;i++) {
    stats[i] = 0;
  }
#endif
  rxof = 0;
#ifdef TSUART_DEBUG
  printk("tsuart_startup called, irq=%d\n",port->irq);
  printk("membase=%p\nmapbase=%lX\n",port->membase,port->mapbase);
  printk("port=%d,NINER=%d\n",port-tsuart_ports,NINER);
#endif
  tmp = request_irq(port->irq, tsuart_int8, 0, "tsuart_uarts", port);
  return tmp;
}

static void tsuart_shutdown(struct uart_port *port)
{
#ifdef TSUART_STATS
  int i;
#endif

  running[port->line] = 0;

  if (--started) {
    return;
  }
  del_timer (&timer);
  timer_added = 0;
#ifdef TSUART_STATS
  for (i=0;i<6;i++) {
    printk("tsuart@%d:%d\n",i+1,stats[i]);
  }
#endif
#ifdef TSUART_DEBUG
  printk("tsuart:overflows=%d\n",rxof);
  printk("tsuart_shutdown called, irq=%d\n",port->irq);
#endif
  put_tsuart_STAT(port,assign_bits_BAUD(get_tsuart_STAT(port),0));

  free_irq(port->irq,port);
  // Since there is no way to disable the UARTs from
  // generating an IRQ, shutdown should probably never be called!
}

static void tsuart_change_speed(struct uart_port *port, unsigned int cflag, unsigned int iflag, unsigned int quot)
{
  int encoded_baud,baud;
#ifdef TSUART_DEBUG
  printk("tsuart_change_speed called, port=%p,cflag=%d,iflag=%d,quot=%d\n",port,cflag,iflag,quot);
#endif	

  encoded_baud = cflag & CBAUD;
  switch (encoded_baud) {
  case B2400:   baud = 2400; break;
  case B4800:   baud = 4800; break;
  case B9600:   baud = 9600; break;
  case B19200:  baud = 19200; break;
  case B38400:  baud = 38400; break;
  case B57600:  baud = 57600; break;
  case B115200: baud = 115200; break;
  default: baud=0;
  }
  if (baud) {
    printk("tsuart:change_speed, port %d: %d baud\n",port->line,baud);
    put_tsuart_STAT(port,assign_bits_BAUD(get_tsuart_STAT(port),baud));
    
    flushing[port->line] = 1;
    udelay(10000);
    flushing[port->line] = 0;
  } else {
    printk("tsuart:unknown change_speed, port %d: %X\n",port->line,encoded_baud);
    //    printk("%08X & %08X -> %08X\n",cflag,CBAUD,cflag & CBAUD);
  }

  if (cflag & PARENB) {
    printk("tsuart:parity unsupported, using no parity\n");
  }
  /* byte size and parity */
  switch (cflag & CSIZE) {
  case CS8:
    break;
  default: // not CS8
    printk("tsuart:unsupported byte size, using CS8\n");
    break;
  }
  if (cflag & CSTOPB) {
    printk("tsuart:unsupported stop bits, using 1\n");
  }
  if (cflag & PARENB) {
    printk("tsuart:no support for parity\n");
  }
  
}

static const char *tsuart_type(struct uart_port *port)
{
  return port->type == PORT_TSUART ? "TSUART" : NULL;
}

/*
 * Release the memory region(s) being used by 'port'
 */
static void tsuart_release_port(struct uart_port *port)
{
  if (port > &tsuart_ports[7]) {
    port->membase = 0;
    return;
  }

#ifdef TSUART_DEBUG
  printk("tsuart_release_port called, membase=%p\n",port->membase);
#endif
  //  release_mem_region(port->mapbase, UART_PORT_SIZE);
  __iounmap(port->membase);

  port->membase = 0;
}

/*
 * Request the memory region(s) being used by 'port'
 */
static int tsuart_request_port(struct uart_port *port)
{
  if (port > &tsuart_ports[7]) {
    struct uart_port *base = &tsuart_ports[(port - &tsuart_ports[8])];
    port->membase = base->membase;
    return 0;
  }
#ifdef TSUART_DEBUG
  printk("tsuart_request_port called, mapbase=%lX\n",port->mapbase);
#endif
  // call __ioremap to avoid a IO to PHY translation - we already have
  // the PHYsical address!
  if (port->membase == 0) {
    port->membase = (void *) __ioremap(
			      port->mapbase & 0xFFFFF000, 4096,0) 
      + (port->mapbase & 0xFFF);
  }
  if (!port->membase) {
    printk(KERN_ERR "tsuart: cannot map io memory\n");
    return -ENOMEM;
  }
  return 0;
  //  return request_mem_region(port->mapbase, UART_PORT_SIZE, "tsuart_uarts")
  //			!= NULL ? 0 : -EBUSY;
}

/*
 * Configure/autoconfigure the port.
 */
static void tsuart_config_port(struct uart_port *port, int flags)
{
  if (flags & UART_CONFIG_TYPE) {
    port->type = PORT_TSUART;
    tsuart_request_port(port);
  }
}

/*
 * verify the new serial_struct (for TIOCSSERIAL).
 */
static int tsuart_verify_port(struct uart_port *port, struct serial_struct *ser)
{
  int ret = 0;
#ifdef TSUART_DEBUG
  printk("tsuart_verify_port\n");
#endif
  if (port == NULL) return -EINVAL; // avoid warning
  if (ser->type != PORT_UNKNOWN && ser->type != PORT_TSUART)
    ret = -EINVAL;
  if (ser->irq < 0 || ser->irq >= NR_IRQS)
    ret = -EINVAL;
  //  if (ser->baud_base != 115200)
  //    ret = -EINVAL;
#ifdef TSUART_DEBUG
  printk("tsuart_verify_port returns %d\n",ret);
#endif
  return ret;
}

static int tsuart_ioctls (struct uart_port *port, unsigned int cmd, unsigned long arg)
{
  return -ENOIOCTLCMD;
}


//-----------------------------------------------------------------------
// Module Definitions                                                   /
//                                                                      \_
//                                                                      /
//                                                                      \_
//                                                                      /
//                                                                      \_
//-----------------------------------------------------------------------



static struct uart_ops tsuart_ops = {
	.stop_tx	= tsuart_stop_tx,
	.start_tx	= tsuart_start_tx,
	.stop_rx	= tsuart_stop_rx,
	.enable_ms	= tsuart_enable_ms,
 	.tx_empty	= tsuart_tx_empty,
	.get_mctrl	= tsuart_get_mctrl,
	.set_mctrl	= tsuart_set_mctrl,
	.break_ctl	= tsuart_break_ctl,
	.startup	= tsuart_startup,
	.shutdown	= tsuart_shutdown,
	.change_speed	= tsuart_change_speed,
	.type		= tsuart_type,
	.release_port	= tsuart_release_port,
	.request_port	= tsuart_request_port,
	.config_port	= tsuart_config_port,
	.verify_port	= tsuart_verify_port,
	.ioctl		= tsuart_ioctls,
};

static struct uart_port tsuart_ports[16] = {
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 0,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+4,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 1,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+8,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 2,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+12,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 3,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+16,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 4,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+20,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 5,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+24,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 6,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+28,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 7,
	},

	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 0,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+4,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 1,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+8,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 2,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+12,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 3,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+16,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 4,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+20,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 5,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+24,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 6,
	},
	{
	  .membase	= 0,
	  .mapbase	= TSUART1_BASE+28,
	  .iotype	= SERIAL_IO_MEM,
	  .irq		= TSUART_IRQ,
	  .uartclk	= UART_CLOCK,
	  .fifosize	= UART_FIFOSIZE,
	  .ops		= &tsuart_ops,
	  .flags	= ASYNC_BOOT_AUTOCONF,
	  .line		= 7,
	}
};

static struct tty_driver normal, callout;
static struct tty_driver normal2, callout2;
static struct tty_struct *cs_tsuart_table[UART_NR];
static struct termios *cs_tsuart_termios[UART_NR], 
			*cs_tsuart_termios_locked[UART_NR];

static struct uart_driver tsuart_reg = {
	.owner			= THIS_MODULE,
	.normal_major		= SERIAL_TSUART_MAJOR,
#ifdef CONFIG_DEVFS_FS
	.normal_name		= SERIAL_TSUART_NAME "%d",
	.callout_name		= CALLOUT_TSUART_NAME "%d",
#else
	.normal_name		= SERIAL_TSUART_NAME,
	.callout_name		= CALLOUT_TSUART_NAME,
#endif
	.normal_driver		= &normal,
	.callout_major		= CALLOUT_TSUART_MAJOR,
	.callout_driver		= &callout,
	.table			= cs_tsuart_table,
	.termios		= cs_tsuart_termios,
	.termios_locked	        = cs_tsuart_termios_locked,
	.minor			= SERIAL_TSUART_MINOR,
	.nr			= UART_NR/2,
	.cons			= NULL,
};

static struct uart_driver tsuart9_reg = {
	.owner			= THIS_MODULE,
	.normal_major		= SERIAL_TSUART_MAJOR,
#ifdef CONFIG_DEVFS_FS
	.normal_name		= SERIAL_TSUART9_NAME "%d",
	.callout_name		= CALLOUT_TSUART9_NAME "%d",
#else
	.normal_name		= SERIAL_TSUART9_NAME,
	.callout_name		= CALLOUT_TSUART9_NAME,
#endif
	.normal_driver		= &normal2,
	.callout_major		= CALLOUT_TSUART_MAJOR,
	.callout_driver		= &callout2,
	.table			= &cs_tsuart_table[8],
	.termios		= &cs_tsuart_termios[8],
	.termios_locked	        = &cs_tsuart_termios_locked[8],
	.minor			= SERIAL_TSUART_MINOR+8,
	.nr			= UART_NR/2,
	.cons			= NULL,
};


static int __init tsuart_init(void)
{
	int ret;
	
	ret = procfile_init();
	if (ret) {
	  return ret;
	}

	/*
	if (get_ts_model() != 2) {
	    printk("tsuart not loading: ts7260 board not detected\n");
	    return -1;
	}
	*/
	ret = uart_register_driver(&tsuart_reg);
	ret |= uart_register_driver(&tsuart9_reg);
	if (ret == 0) {
		int i;

		for (i=0;i<UART_NR;i++) {
 		        flushing[i] = 0;
			running[i] = 0;
			tsuart_ports[i].unused[0] = 1;
			tsuart_ports[i].unused[1] = 1;
			tsuart_ports[i].unused[4] = (i > 7);
			uart_add_one_port(i < 8 ? &tsuart_reg:&tsuart9_reg,
					  &tsuart_ports[i]);
		}
	}
	return ret;
}

static void __exit tsuart_exit(void)
{
	int i;

	procfile_exit();
	for (i=0;i<UART_NR;i++) {
		uart_remove_one_port(i < 8 ? &tsuart_reg:&tsuart9_reg, 
				     &tsuart_ports[i]);
	}
	uart_unregister_driver(&tsuart9_reg);
	uart_unregister_driver(&tsuart_reg);
}

module_init(tsuart_init);
module_exit(tsuart_exit);

EXPORT_NO_SYMBOLS;

MODULE_AUTHOR("Technologic Systems");
MODULE_DESCRIPTION("Technologic Systems UART serial port driver");
MODULE_LICENSE("GPL");  // or BSD.  Take your pick!
// agcc -W -Wall -O2 -c -D__KERNEL__ -DMODULE -isystem ../../include tsuart.c
