/*****************************************************************************
 * linux/drivers/video/ts7kvfb.c -- TS7KV frame buffer device
 *
 *  Created 24 Aug 2005 by Eddie Dawydiuk
 *
 *  Based on:
 *   linux/drivers/video/skeletonfb.c
 *     Created 28 Dec 1997 by Geert Uytterhoeven
 *
 *   linux/drivers/video/au1100fb.c
 *     Author: MontaVista Software, Inc.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive
 * for more details.
 ****************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <video/fbcon.h>
#include <video/fbcon-cfb16.h>
#include "ts7kvfb.h"

#ifdef CONFIG_LOAD_FPGA
#include "load_fpga.c"
#endif

struct ts7kvfb_info {
	struct fb_info_gen gen;
	unsigned long fb_virt_start;
	unsigned long fb_size;
	unsigned long fb_phys;
	volatile unsigned short *fpga_regs;
	unsigned int page;
	struct { unsigned red, green, blue, pad; } palette[16];
	int xres;
	int yres;
	char id;

	u16 fbcon_cmap16[16];
	struct display_switch dispw;

};

struct ts7kvfb_par {
	struct fb_var_screeninfo var;
};

static struct ts7kvfb_info fb_info;
static struct ts7kvfb_par current_par;
static struct display disp;

int ts7kvfb_init(void);
int ts7kvfb_setup(char *options);
static int ts7kv_blank(int blank_mode, struct fb_info_gen *info);

static struct fb_ops ts7kvfb_ops = {
	owner:		THIS_MODULE,
	fb_get_fix:	fbgen_get_fix,
	fb_get_var:	fbgen_get_var,
	fb_set_var:	fbgen_set_var,
	fb_pan_display: fbgen_pan_display,
	fb_get_cmap:	fbgen_get_cmap,
	fb_set_cmap:	fbgen_set_cmap,
	fb_pan_display:	fbgen_pan_display,
};

void ts7kv_blt(int src, int dest, int width, int height, int pxfill)
{
	unsigned short swin = fb_info.fpga_regs[DEVSEL];
	unsigned short vidctrl;
	unsigned int direction = 0;

	if (pxfill == 0 && src < dest) {
		src += (height - 1) * fb_info.xres;
		dest += (height - 1) * fb_info.xres;
		direction = 1;
	}

	fb_info.fpga_regs[DEVSEL] = 0x50;
	// XXX: This should probably be a read-modify-write.  --JO
	vidctrl = fb_info.fpga_regs[VIDCTRL];
	fb_info.fpga_regs[VIDCTRL] = (vidctrl & 0xff3f) | (direction << 7) | (pxfill << 6);
	fb_info.fpga_regs[BLTCTRL] = (src >> 16) | ((dest >> 16) << 6) | ((width >> 7) << 13);
	fb_info.fpga_regs[BLTSZ] = (height & 0x1ff) | (width << 9);
	fb_info.fpga_regs[SRCBLT] = src & 0xffff;
	fb_info.fpga_regs[DSTBLT] = dest & 0xffff;
	while (fb_info.fpga_regs[VIDCTRL] & 0x400);

	fb_info.fpga_regs[DEVSEL] = swin;
}


static void fbcon_ts7kv_bmove(struct display *p, int sy, int sx, int dy, int dx,
	                        int height, int width)
{
	unsigned int src = sy * fontheight(p) * fb_info.xres + sx * fontwidth(p);
	unsigned int dst = dy * fontheight(p) * fb_info.xres + dx * fontwidth(p);

	width  *= fontwidth(p);
	height *= fontheight(p);

	ts7kv_blt(src, dst, width - 1, height, 0);
}

void fbcon_ts7kv_clear(struct vc_data *conp, struct display *p, int sy, int sx,
                       int height, int width)
{
	unsigned int dst = sy * fontheight(p) * fb_info.xres + sx * fontwidth(p);
	unsigned short bgcolor;

	width  *= fontwidth(p);
	height *= fontheight(p);
	bgcolor = ((u16 *)p->dispsw_data)[attr_bgcol_ec(p, conp)];

	ts7kv_blt(bgcolor, dst, width - 1, height, 1);
}


struct display_switch fbcon_ts7kv = {
    setup:              fbcon_cfb16_setup,
    bmove:              fbcon_ts7kv_bmove,
    clear:              fbcon_ts7kv_clear,
    putc:               fbcon_cfb16_putc,
    putcs:              fbcon_cfb16_putcs,
    revc:               fbcon_cfb16_revc,
    clear_margins:      fbcon_cfb16_clear_margins,
    fontwidthmask:      FONTWIDTH(4)|FONTWIDTH(8)|FONTWIDTH(12)|FONTWIDTH(16)
};

static void ts7kv_detect(void)
{
}

static int ts7kv_encode_fix(struct fb_fix_screeninfo *fix,
				const void *_par, struct fb_info_gen *_info)
{
	struct ts7kvfb_info *info = (struct ts7kvfb_info *) _info;

	memset(fix, 0, sizeof(struct fb_fix_screeninfo));

	fix->smem_start = info->fb_phys;
	fix->smem_len = info->fb_size;
	fix->type = FB_TYPE_PACKED_PIXELS;
	fix->type_aux = 0;
	fix->visual = FB_VISUAL_TRUECOLOR;
	fix->ywrapstep = 0;
	fix->xpanstep = 0;
	fix->ypanstep = 0;
	fix->line_length = fb_info.xres * 2;
	fix->accel = FB_ACCELF_TEXT;

	return 0;
}

static void set_color_bitfields(struct fb_var_screeninfo *var)
{
	var->red.offset = 11;
	var->green.offset = 6;
	var->blue.offset = 0;
	var->transp.offset = 0;

	var->red.length = 5;
	var->green.length = 6;
	var->blue.length = 5;
	var->transp.length = 0;

	var->red.msb_right = 0;
	var->green.msb_right = 0;
	var->blue.msb_right = 0;
	var->transp.msb_right = 0;
}

static int ts7kv_decode_var(const struct fb_var_screeninfo *var,
				void *_par, struct fb_info_gen *_info)
{
	struct ts7kvfb_par *par = (struct ts7kvfb_par *)_par;

	if (var->xres != fb_info.xres || var->yres != fb_info.yres)
		return -EINVAL;

	if(var->bits_per_pixel != BPP)
		return -EINVAL;

	memset(par, 0, sizeof(struct ts7kvfb_par));
	par->var = *var;
	par->var.bits_per_pixel = 16;
	set_color_bitfields(&par->var);

	return 0;
}

static int ts7kv_encode_var(struct fb_var_screeninfo *var,
	const void *par, struct fb_info_gen *info)
{
	*var = ((struct ts7kvfb_par *)par)->var;
	return 0;
}

static void ts7kv_get_par(void *_par, struct fb_info_gen *_info)
{
	*(struct ts7kvfb_par *)_par = current_par;
}

static void ts7kv_set_par(const void *par, struct fb_info_gen *info)
{
	/* nothing to do: we don't change any settings */
}

static int ts7kv_getcolreg(unsigned regno, unsigned *red, unsigned *green,
			 unsigned *blue, unsigned *transp,
			 struct fb_info *info)
{
	struct ts7kvfb_info* i = (struct ts7kvfb_info*)info;

	if (regno > 15)
		return 1;

	*red    = i->palette[regno].red;
	*green  = i->palette[regno].green;
	*blue   = i->palette[regno].blue;
	*transp = 0;

	return 0;
}

static int ts7kv_setcolreg(unsigned regno, unsigned red, unsigned green,
			 	unsigned blue, unsigned transp,
				struct fb_info *info)
{
	struct ts7kvfb_info* i = (struct ts7kvfb_info *)info;

	if (regno > 15)
		return 1;

	i->palette[regno].red    = red;
	i->palette[regno].green  = green;
	i->palette[regno].blue   = blue;

	i->fbcon_cmap16[regno] =
		((red & 0xf800) >> 0) |
		((green & 0xf800) >> 5) |
		((blue & 0xf800) >> 11);

	return 0;
}

static int ts7kv_pan_display(const struct fb_var_screeninfo *var,
				struct fb_info_gen *info)
{
	return 0;
}

/*
 *  Blank the screen if blank_mode != 0, else unblank. If blank == NULL
 *  then the caller blanks by setting the CLUT (Color Look Up Table) to all
 *  black. Return 0 if blanking succeeded, != 0 if un-/blanking failed due
 *  to e.g. a video mode which doesn't support it. Implements VESA suspend
 *  and powerdown modes on hardware that supports disabling hsync/vsync:
 *    blank_mode == 2: suspend vsync
 *    blank_mode == 3: suspend hsync
 *    blank_mode == 4: powerdown
 */
static int ts7kv_blank(int blank_mode, struct fb_info_gen *info)
{
	unsigned short swin = fb_info.fpga_regs[DEVSEL];

	if(fb_info.id == LCD_MONO)
		return 0;

	fb_info.fpga_regs[DEVSEL] = 0x50;
	switch (blank_mode) {
	case 0:
		fb_info.fpga_regs[VIDCTRL] = 0x300;
		break;
	case 1:
	case 2:
		fb_info.fpga_regs[VIDCTRL] = 0x200;
		break;
	case 3:
		fb_info.fpga_regs[VIDCTRL] = 0x100;
		break;
	case 4:
		fb_info.fpga_regs[VIDCTRL] = 0x0;
		break;
	}

	fb_info.fpga_regs[DEVSEL] = swin;

	return 0;
}

static void ts7kv_set_disp(const void *unused, struct display *disp,
			 struct fb_info_gen *info)
{
	disp->screen_base = (char *)fb_info.fb_virt_start;
	disp->dispsw = &fbcon_ts7kv;
	disp->dispsw_data = fb_info.fbcon_cmap16;
	disp->scrollmode = 0;
}

static struct fbgen_hwswitch ts7kv_switch =
{
	ts7kv_detect,
	ts7kv_encode_fix,
	ts7kv_decode_var,
	ts7kv_encode_var,
	ts7kv_get_par,
	ts7kv_set_par,
	ts7kv_getcolreg,
	ts7kv_setcolreg,
	ts7kv_pan_display,
	ts7kv_blank,
	ts7kv_set_disp
};

int __init ts7kvfb_init(void)
{
	int return_val = 0;
	unsigned long base_addr;
	int i;

#ifdef CONFIG_LOAD_FPGA
	//Load the firmware
	init_fpga();
	return_val = write_bitstream();
	if(return_val != 0) {
		printk("Error loading firmware...\n");
		return -1;
	}
#endif

	fb_info.fb_size = VIDEOMEMORYSIZE;
	for (i=0;i<5;i++)
	{
		if (i==4) {
			printk("TS7KV board was not found or firmware not loaded!\n");
			return -1;
		}
		base_addr = TS7XXX_IOMEM|BASE_ADDR + (i*BASE_SIZE);
		if ( (inb(base_addr|ID0_REG) == TS7KV_ID0) &&
			(inb(base_addr|ID1_REG) == TS7KV_ID1) &&
			((inb(base_addr|CTR_REG0))&DONE_MSK == DONE_MSK) )
		{
			fb_info.fb_virt_start = IO_ADDRESS(TS7XXX_IO16_BASE + 0x1800000 + (i*VIDEOMEMORYSIZE));
			if (check_mem_region(fb_info.fb_virt_start, fb_info.fb_size))
				continue;
			request_mem_region(fb_info.fb_virt_start, fb_info.fb_size, "TS-7KV");
			fb_info.fb_phys = 0x21800000 + (i*VIDEOMEMORYSIZE);
			fb_info.fpga_regs = (short *)IO_ADDRESS(TS7XXX_IO16_BASE +
				TS7XXX_ISAIO + REGS_ADDR + (i*REGS_SIZE));
			printk(KERN_INFO "Framebuffer at 0x%lx allocated for TS-7KV with jumpers #1:%s #2:%s\n",
				fb_info.fb_phys, ((base_addr&0x8) ? "ON" : "OFF"), ((base_addr&0x10) ? "ON" : "OFF"));
			break;
		}
	}

	unsigned short swin = fb_info.fpga_regs[DEVSEL];
	fb_info.fpga_regs[DEVSEL] = 0x50;

	if(fb_info.fpga_regs[BLTSZ] == CRT) 
	{
		fb_info.xres = 640;
		fb_info.yres = 320;
		fb_info.id = CRT;
	}else if(fb_info.fpga_regs[BLTSZ] == LCD_MONO) 
	{
		fb_info.xres = 320;
		fb_info.yres = 240;
		fb_info.id = LCD_MONO;
	}else
	{
		printk("This version doesn't support the current bitstream.\n");
		return -1;
	}		

	current_par.var.xres = fb_info.xres;
	current_par.var.xres_virtual = fb_info.xres;
	current_par.var.yres = fb_info.yres;
	current_par.var.yres_virtual = fb_info.yres;
	current_par.var.bits_per_pixel = BPP;

	//timing information
	current_par.var.pixclock = 31746;    //picoseconds
	current_par.var.left_margin =  128;  //pixels, (back porch)
	current_par.var.right_margin = 24;   //pixels, (front porch)
	current_par.var.upper_margin = 28;   //pixel lines, (back porch)
	current_par.var.lower_margin = 9;    //pixel lines, (front porch)
	current_par.var.hsync_len = 40;      //pixels
	current_par.var.vsync_len = 3;       //pixel lines

	udelay(350);

	{
		fb_info.fpga_regs[VIDCTRL] = 0x200;
		mdelay(100);
		fb_info.fpga_regs[VIDCTRL] = 0x300;
		fb_info.fpga_regs[DEVSEL] = swin;
		fb_info.page = 0;
	}

	// Blank screen
	ts7kv_blt(0, 0, 639, 480, 1);

	strcpy(fb_info.gen.info.modename, "TS-7KV");
	strcpy(fb_info.gen.info.fontname, "Acorn8x8");
	fb_info.gen.parsize = sizeof(struct ts7kvfb_par);
	fb_info.gen.fbhw = &ts7kv_switch;
	fb_info.gen.info.changevar = NULL;
	fb_info.gen.info.node = -1;
	fb_info.gen.info.fbops = &ts7kvfb_ops;
	fb_info.gen.info.disp = &disp;
	fb_info.gen.info.switch_con = &fbgen_switch;
	fb_info.gen.info.updatevar = &fbgen_update_var;
	fb_info.gen.info.blank = &fbgen_blank;
	fb_info.gen.info.flags = FBINFO_FLAG_DEFAULT;

	/* This should give a reasonable default video mode */
	fbgen_get_var(&disp.var, -1, &fb_info.gen.info);
	fbgen_do_set_var(&disp.var, 1, &fb_info.gen);
	fbgen_set_disp(-1, &fb_info.gen);

	if (register_framebuffer(&fb_info.gen.info) < 0)
		return -EINVAL;

	printk(KERN_INFO "fb%d: %s frame buffer device\n",
		GET_FB_IDX(fb_info.gen.info.node), fb_info.gen.info.modename);

	return return_val;
}

int __init ts7kvfb_setup(char *options)
{
	/* Parse user speficied options (`video=ts7kvfb:') */
	return 0;
}

#ifdef MODULE
MODULE_LICENSE("GPL");
int init_module(void)
{
    return ts7kvfb_init();
}

void cleanup_module(void)
{
	unregister_framebuffer(&fb_info);
	release_mem_region(fb_info.fb_virt_start, fb_info.fb_size);
}
#endif /* MODULE */
